unit DBF;

interface

uses
  Classes, Windows, SysUtils;

////////////////////////////////////////////////////////////////////////////////
//
// 00         FoxBase+/dBase no memo                         03H
//            FoxBase+/dBase with memo                       83H
//            FoxPro/dBase4 no memo                          03H
//            FoxPro with memo                               F5H
//            dBase4 with mem0                               8Bh
//            Visual FoxPro                                  30h
// ---------------------------------------------------------------
// 01-03      YY MM DD
// ---------------------------------------------------------------
// 04-07      Number of records (dword)
// ---------------------------------------------------------------
// 08-09      Position of first data record
// ---------------------------------------------------------------
// 10-11      Length of one data record (including delete flag)
// ---------------------------------------------------------------
// 12-31      Reserved, but if byte 28 = 01 - with CDX
//                          if byte 28 = 02 - with memo file (VFP)
//            Also byte 29 could be a codepage number
// ---------------------------------------------------------------
// 32-n       Fields
// ---------------------------------------------------------------
// n+1        Header terminator (0DH)
// ===============================================================
//
////////////////////////////////////////////////////////////////////////////////
//
// Field structure
//
// ===============================================================
// 00-10      Field name (ASCIIZ, padded with 00)
// ---------------------------------------------------------------
// 11         Type:
//              C - Character
//              N - Numeric
//              L - Logical
//              M - Memo
//              G - General
//              D - Date        (YYYYMMDD)
//              F - Float
//              P - Picture
// ---------------------------------------------------------------
// 12-15      Offset of the field within a record
// ---------------------------------------------------------------
// 16         Length of field
// ---------------------------------------------------------------
// 17         Number of decimal places
// ---------------------------------------------------------------
// 18-31      Reserved
//
////////////////////////////////////////////////////////////////////////////////

type
  TDBFHeader = record
    FileType     : byte;
    YY           : byte;
    MM           : byte;
    DD           : byte;
    RecCount     : dword;
    DataPosition : word;
    RecLength    : word;
    Reserved1    : array [12..27] of byte;
    Flags        : byte;
    Language     : byte;
    Reserved2    : array [30..31] of byte;
  end;

type
  PDBFField = ^TDBFField;
  TDBFField = record
    FieldName      : array [0..10] of char;
    FieldType      : char;
    FieldOffset    : dword;
    FieldSize      : byte;
    FieldPrecision : byte;
    Reserved3      : array [18..31] of byte;
  end;

type
  TDBFFieldDef = packed record
    FieldName   : string[20];
    FieldOffset : integer;
    FieldType   : char;
    FieldSize   : integer;
    Precision   : integer;
  end;

type
  TDBF = class (TList)
    constructor Create(AFileName : string);
    destructor Destroy; override;
    procedure Clear; override;
  private
    FFileName : string;
    FStream   : TFileStream;
    FHDR      : TDBFHeader;
    FBuffer   : pchar;
    procedure LoadHeaders;
    function GetFieldDef(Idx : integer) : TDBFFieldDef;
    function GetFieldValue(Idx : integer) : string;
    function GetLanguageCode : byte;
  public
    RecNo     : dword;
    EOF       : boolean;
    property Language  : byte read GetLanguageCode;
    property FieldDefs[Idx : integer] : TDBFFieldDef read GetFieldDef; default;
    property Fields[Idx : integer] : string read GetFieldValue;
    function FieldByName(FieldName : string) : string;
    procedure Next;
    procedure Prev;
    procedure GoToRec(ARecNo : dword);
  end;

implementation

// ========== TDBF =============================================================

constructor TDBF.Create(AFileName : string);
begin
  FFileName := AFileName;
  FStream   := TFileStream.Create(FFileName,fmOpenRead+fmShareDenyNone);
  FBuffer   := nil;
  EOF       := true;
  RecNo     := 1;
  LoadHeaders;
end;

// -----------------------------------------------------------------------------

destructor TDBF.Destroy;
begin
  FStream.Free;
end;

// -----------------------------------------------------------------------------

procedure TDBF.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
    FreeMem(PDBFField(items[i]));
  if assigned(FBuffer) then FreeMem(FBuffer);
  FBuffer := nil;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

procedure TDBF.LoadHeaders;
var
  fld  : TDBFField;
  pfld : PDBFField;
begin
  if FStream.Read(FHDR,sizeof(TDBFHeader)) <> sizeof(TDBFHeader) then
    raise Exception.Create('Incorrect header length');
  if FHDR.RecLength > 10000 then
    raise Exception.Create('Record is too long');
  if assigned(FBuffer) then FreeMem(FBuffer);
  FBuffer := nil;
  repeat
    if FStream.Read(fld,SizeOf(TDBFField)) <> SizeOf(TDBFField) then break;
    if (fld.FieldName[0] = #13) or (FStream.Position > FHDR.DataPosition) then break;
    GetMem(pfld,sizeof(TDBFField));
    pfld^ := fld;
    Add(pfld);
  until false;
  FStream.Seek(FHDR.DataPosition,soFromBeginning);
  GetMem(FBuffer,FHDR.RecLength);
  EOF := (FStream.Read(FBuffer^,FHDR.RecLength) <> FHDR.RecLength);
  if EOF then
    RecNo := 0
  else
    RecNo := 1;
end;

// -----------------------------------------------------------------------------

function TDBF.GetFieldDef(Idx : integer) : TDBFFieldDef;
var
  i : integer;
  p : PDBFField;
begin
  if (Idx < 0) or (Idx >= Count) then
  begin
    result.FieldName := '';
    result.FieldOffset := 0;
    result.FieldType := #0;
    result.FieldSize := 0;
    result.Precision := 0;
  end
  else
  begin
    p := items[Idx];
    for i:=0 to 10 do
      if p.FieldName[i] <> #0 then
      begin
        result.FieldName[0] := chr(i+1);
        result.FieldName[i+1] := p.FieldName[i];
      end;
    result.FieldOffset := p.FieldOffset;
    result.FieldType   := p.FieldType;
    result.FieldSize   := p.FieldSize;
    result.Precision   := p.FieldPrecision;
  end;
end;

// -----------------------------------------------------------------------------

function TDBF.FieldByName(FieldName : string) : string;
var
  ii : integer;
begin
  result := '';
  FieldName := uppercase(FieldName);
  for ii:=0 to Count-1 do
  begin
    if uppercase(FieldDefs[ii].FieldName) = FieldName then
    begin
      result := GetFieldValue(ii);
      exit;
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TDBF.GetLanguageCode : byte;
begin
  result := FHDR.Language;
end;

function TDBF.GetFieldValue(Idx : integer) : string;
var
  i : integer;
  j : integer;
  d : TDBFFieldDef;
  o : integer;
  s : string;
  dd,mm,yy : word;
begin
  result := '';
  if not assigned(FBuffer) then exit;
  d := FieldDefs[Idx];
  o := 1;
  for i:=0 to Idx-1 do o := o+FieldDefs[i].FieldSize;
  case d.FieldType of
    'C':
    begin
      setlength(result,d.FieldSize);
      for i:=0 to d.FieldSize-1 do result[i+1] := FBuffer[o+i];
      i := d.FieldSize;
      while (i > 0) and (result[i] in [' ',#0]) do dec(i);
      setlength(result,i);
    end;
    'N':
    begin
      i := 0;
      while (i < d.FieldSize) and (FBuffer[i+o] in [' ',#0]) do inc(i);
      if i <= d.FieldSize then
      begin
        setlength(result,d.FieldSize-i);
        j := 1;
        while (i < d.FieldSize) do
        begin
          result[j] := FBuffer[o+i];
          inc(i);
          inc(j);
        end;
      end;
    end;
    'L':
    begin
      result := FBuffer[o];
      if result = 'T' then
        result := 'TRUE'
      else
        result := 'FALSE';
    end;
    'D':
    begin
      setlength(s,8);
      copymemory(@s[1],@FBuffer[o],8);
      yy := StrToIntDef(copy(s,1,4),2005);
      mm := StrToIntDef(copy(s,5,2),12);
      dd := StrToIntDef(copy(s,7,2),22);
      result := format('%2.2u/%2.2u/%4.4u',[mm,dd,yy]);
    end;
    'F':
    begin
      i := 0;
      while (i < d.FieldSize) and (FBuffer[i+o] in [' ',#0]) do inc(i);
      if i <= d.FieldSize then
      begin
        setlength(result,d.FieldSize-i);
        j := 1;
        while (i < d.FieldSize) do
        begin
          result[j] := FBuffer[o+i];
          inc(i);
          inc(j);
        end;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------

procedure TDBF.Next;
var
  position : integer;
begin
  if RecNo < FHDR.RecCount then
  begin
    inc(RecNo);
    position := FHDR.DataPosition+((RecNo-1)*FHDR.RecLength);
    FStream.Position := Position;
    EOF := (FStream.Read(FBuffer^,FHDR.RecLength) <> FHDR.RecLength);
  end
  else
    EOF := true;
end;

// -----------------------------------------------------------------------------

procedure TDBF.Prev;
var
  position : integer;
begin
  if RecNo > 1 then
  begin
    dec(RecNo);
    position := FHDR.DataPosition+((RecNo-1)*FHDR.RecLength);
    FStream.Position := Position;
    EOF := (FStream.Read(FBuffer^,FHDR.RecLength) <> FHDR.RecLength);
  end;
end;

// -----------------------------------------------------------------------------

procedure TDBF.GoToRec(ARecNo : dword);
var
  position : integer;
begin
  if (ARecNo > 0) and (ARecNo <= FHDR.RecCount) then
  begin
    RecNo := ARecNo;
    position := FHDR.DataPosition+((RecNo-1)*FHDR.RecLength);
    FStream.Position := Position;
    EOF := (FStream.Read(FBuffer^,FHDR.RecLength) <> FHDR.RecLength);
  end;
end;

// =============================================================================

end.
