unit ESRI;

interface

uses
  Classes, Types, ShapeFileObjects, Windows;

type
  TCharset = record
    CharsetString   : string;
    DbfLanguageCode : byte;
    CharsetCode     : byte;
  end;

const
  MAX_CP = 24;

var
  CharsetTable : array [1..MAX_CP] of TCharset = (
      (CharsetString:  '437'; DbfLanguageCode: $01; CharsetCode: ANSI_CHARSET),        // U.S. MS-DOS
      (CharsetString:  '737'; DbfLanguageCode: $6a; CharsetCode: ANSI_CHARSET),        // Greek MS-DOS (437G)
      (CharsetString:  '850'; DbfLanguageCode: $02; CharsetCode: ANSI_CHARSET),        // International MS-DOS
      (CharsetString:  '852'; DbfLanguageCode: $64; CharsetCode: ANSI_CHARSET),        // Czech OEM
      (CharsetString:  '857'; DbfLanguageCode: $6b; CharsetCode: ANSI_CHARSET),        // Turkish MS-DOS
      (CharsetString:  '860'; DbfLanguageCode: $00; CharsetCode: ANSI_CHARSET),        // Portugese OEM
      (CharsetString:  '861'; DbfLanguageCode: $67; CharsetCode: ANSI_CHARSET),        // Icelandic MS-DOS
      (CharsetString:  '863'; DbfLanguageCode: $00; CharsetCode: ANSI_CHARSET),        // French OEM (Canada)
      (CharsetString:  '865'; DbfLanguageCode: $65; CharsetCode: ANSI_CHARSET),        // Danish OEM
      (CharsetString:  '866'; DbfLanguageCode: $66; CharsetCode: ANSI_CHARSET),        // Russian OEM
      (CharsetString:  '874'; DbfLanguageCode: $00; CharsetCode: THAI_CHARSET),        // Thai (ANSI/OEM)
      (CharsetString:  '932'; DbfLanguageCode: $7b; CharsetCode: SHIFTJIS_CHARSET),    // Japanese Shift-JIS
      (CharsetString:  '936'; DbfLanguageCode: $7a; CharsetCode: ANSI_CHARSET),        // Chinese GBK (PRC)
      (CharsetString:  '949'; DbfLanguageCode: $00; CharsetCode: ANSI_CHARSET),        // Korean (ANSI/OEM)
      (CharsetString:  '950'; DbfLanguageCode: $78; CharsetCode: CHINESEBIG5_CHARSET), // Chinese Big 5 (Taiwan)
      (CharsetString: '1250'; DbfLanguageCode: $c8; CharsetCode: EASTEUROPE_CHARSET),  // Eastern European Windows
      (CharsetString: '1251'; DbfLanguageCode: $c9; CharsetCode: RUSSIAN_CHARSET),     // Russian Windows
      (CharsetString: '1252'; DbfLanguageCode: $57; CharsetCode: ANSI_CHARSET),        // ANSI
      (CharsetString: '1252'; DbfLanguageCode: $03; CharsetCode: ANSI_CHARSET),        // Windows ANSI
      (CharsetString: '1253'; DbfLanguageCode: $cb; CharsetCode: GREEK_CHARSET),       // Greek Windows
      (CharsetString: '1254'; DbfLanguageCode: $ca; CharsetCode: TURKISH_CHARSET),     // Turkish Windows
      (CharsetString: '1255'; DbfLanguageCode: $7d; CharsetCode: HEBREW_CHARSET),      // Hebrew Windows
      (CharsetString: '1256'; DbfLanguageCode: $7e; CharsetCode: ARABIC_CHARSET),      // Arabic Windows
      (CharsetString: '1257'; DbfLanguageCode: $00; CharsetCode: BALTIC_CHARSET));     // Baltic Windows


type
  TDoubleRect = packed record
    Left   : double;
    Bottom : double;
    Right  : double;
    Top    : double;
  end;

type
  TDoublePoint = packed record
    X : double;
    Y : double;
  end;

type
  TDoublePointArray = array [0..$7fffff] of TDoublePoint;
  PDoublePointArray = ^TDoublePointArray;

type
  TESRIFileHeader = packed record
    FileCode    : integer;                  // Big endian
    Unused      : array [1..5] of integer;  // Big endian
    FileLength  : integer;                  // Big endian, in 16-bit words including this header
    FileVersion : integer;
    ShapeType   : integer;
    Rect        : TDoubleRect;
    Zmin        : double;
    Zmax        : double;
    Mmin        : double;
    Mmax        : double;
  end;

const
  ESRINullShape   =  0;
  ESRIPoint       =  1;
  ESRIPolyLine    =  3;
  ESRIPolygon     =  5;
  ESRIMultiPoint  =  8;
  ESRIPointZ      = 11;
  ESRIPolyLineZ   = 13;
  ESRIPolygonZ    = 15;
  ESRIMultiPointZ = 18;
  ESRIPointM      = 21;
  ESRIPolyLineM   = 23;
  ESRIPolygonM    = 25;
  ESRIMultiPointM = 28;
  ESRIMultiPatch  = 31;

type
  TESRIRecordHeader = packed record
    RecordNumber  : integer; // Big endian, 1-based
    ContentLength : integer; // Big endian, in 16-bit words NOT including this header
  end;

type
  TPolyRec = class
    destructor Destroy; override;
  public
    Rect   : TRect;
    Attr   : integer;
    Tag    : integer;
    Parts  : array of integer;
    Points : array of TPoint;
  end;

type
  TESRIShapeList = class (TList)
    procedure Clear; override;
  private
    function GetShape(ShapeIdx : integer) : TPolyRec;
    function GetPolyCount(ShapeIdx : integer) : integer;
    function GetPoly(ShapeIdx : integer; PolyIdx : integer) : TPolygon;
    function GetAttr(ShapeIdx : integer) : integer;
    procedure SetAttr(ShapeIdx : integer; Attr : integer);
  public
    property PolyCount[ShapeIdx : integer] : integer read GetPolyCount;
    property Polys[ShapeIdx : integer; PolyIdx : integer] : TPolygon read GetPoly;
    property Attribute[ShapeIdx : integer] : integer read GetAttr write SetAttr;
    property Shapes[ShapeIdx : integer] : TPolyRec read GetShape; default;
    procedure AddShape(Attr,PartCount,PointCount : integer; Rect : TDoubleRect;  Parts : PIntegerArray; Points : PDoublePointArray);
    function GetShapeByAttr(Attr : integer) : integer;
  end;

type
  TESRILineList = class (TList)
    procedure Clear; override;
  private
    function GetLine(LineIdx : integer) : TPolyRec;
    function GetPoly(LineIdx,PolyIdx : integer) : TPolygon;
    function GetPolyCount(LineIdx : integer) : integer;
  public
    property Lines[LineIdx : integer] : TPolyRec read GetLine; default;
    property Polys[LineIdx : integer; PolyIdx : integer] : TPolygon read GetPoly;
    property PolyCount[LineIdx : integer] : integer read GetPolyCount;
    procedure AddLine(Attr,PartCount,PointCount : integer; Rect : TDoubleRect;  Parts : PIntegerArray; Points : PDoublePointArray);
    procedure SortByAttribute;
  end;

type
  TPointRec = class
    Attr : integer;
    X    : integer;
    Y    : integer;
  end;

type
  TESRIPointList = class (TList)
    procedure Clear; override;
  private
    function GetPointRec(PointIdx : integer) : TPointRec;
  public
    property Points[PointIdx : integer] : TPointRec read GetPointRec; default;
    procedure AddPoint(AttrIdx : integer; Point : TDoublePoint);
    procedure AddMultiplePoints(Attr,PointCount : integer; Points : PDoublePointArray);
  end;

type
  TESRIAttributeValue = class
    constructor Create(ASValue : string);
  public
    SValue   : string;
    NValue   : double;
    DValue   : TDateTime;
    BValue   : boolean;
  end;

type
  TESRIAttribute = class (TList)
    constructor Create(AName : string; ADataType : char);
    procedure Clear; override;
  private
    function GetAttributeValue(Idx : integer) : TESRIAttributeValue;
  public
    Name     : string;
    DataType : char;
    property Values[Idx : integer] : TESRIAttributeValue read GetAttributeValue; default;
    function GetAttrIdxByValue(Value : string) : integer; // For now it works only for strings
    function GetNextAttrIdxByValue(LastIdx : integer; Value : string) : integer; // For now it works only for strings
  end;

type
  TESRIAttributes = class (TList)
    procedure Clear; override;
  private
    function GetAttribute(Idx : integer) : TESRIAttribute;
    function GetRowCount : integer;
  public
    property Fields[Idx : integer] : TESRIAttribute read GetAttribute; default;
    property RowCount : integer read GetRowCount;
  end;

type
  TESRIDataSet = class
    constructor Create(ASHPFileName : string);
    destructor Destroy; override;
  private
    FSHPFileName    : string;
    FDBFFileName    : string;
    FCPGFileName    : string;
    FHDR            : TESRIFileHeader;
    FESRIShapes     : TESRIShapeList;
    FESRILines      : TESRILineList;
    FESRIPoints     : TESRIPointList;
    FESRIAttributes : TESRIAttributes;
    FDBF            : boolean;
    FCodePage       : byte;
    function GetVersion : integer;
    function GetLength : integer;
    function GetShapeType : integer;
    function LoadShpFile : boolean;
    function LoadDbfFile : boolean;
    function GetRect : TRect;
    function GetDBFCodePage(LanguageCode : byte) : byte;
    function GetCPGCodePage(CPG : string) : byte;
  public
    property CodePage    : byte            read FCodePage;
    property DBF         : boolean         read FDBF;
    property SHPFileName : string          read FSHPFileName;
    property DBFFileName : string          read FDBFFileName;
    property Version     : integer         read GetVersion;
    property Length      : integer         read GetLength;
    property ShapeType   : integer         read GetShapeType;
    property Rect        : TRect           read GetRect;
    property Shapes      : TESRIShapeList  read FESRIShapes;
    property Lines       : TESRILineList   read FESRILines;
    property Points      : TESRIPointList  read FESRIPoints;
    property Attributes  : TESRIAttributes read FESRIAttributes;
  end;

function GetESRIPoly(Shape : TPolyRec; PartIdx : integer) : TPolygon;
function GetAttrIdx(ESRIDS : TESRIDataSet; AttrName : string) : integer;

implementation

uses
  SysUtils, DBF, Math;

// =============================================================================

function GetAttrIdx(ESRIDS : TESRIDataSet; AttrName : string) : integer;
var
  ii : integer;
begin
  result := -1;
  if not ESRIDS.DBF then exit;
  AttrName := uppercase(AttrName);
  for ii:=0 to ESRIDS.Attributes.Count-1 do
    if uppercase(ESRIDS.Attributes[ii].Name) = AttrName then
    begin
      result := ii;
      exit;
    end;
end;

// -----------------------------------------------------------------------------

function GetESRIPoly(Shape : TPolyRec; PartIdx : integer) : TPolygon;
var
  Idx,i,j : integer;
  PolyCount : integer;
  p : TPoint;
begin
  PolyCount := length(Shape.Parts);
  if (PartIdx < 0) or (PartIdx >= PolyCount) then
  begin
    setlength(result,0);
    exit;
  end;
  Idx := Shape.Parts[PartIdx];
  p := Shape.Points[Idx];
  j := 1;
  for i:=Idx+1 to length(Shape.Points)-1 do
  begin
    inc(j);
    if (p.X = Shape.Points[i].X) and (p.Y = Shape.Points[i].Y) then break;
  end;
  setlength(result,j);
  for i:=Idx to (Idx+j-1) do
  begin
    result[i-Idx] := Shape.Points[i];
  end;
end;

// -----------------------------------------------------------------------------

function  ReplaceExtension(fn,ext : string) : string;
begin
  result := fn;
  while (result <> '') and (result[length(result)] <> '.') do setlength(result,length(result)-1);
  if result = '' then
    result := fn+ext
  else
    result := result+ext;
end;

// ========== TESRIAttributeValue ==============================================

constructor TESRIAttributeValue.Create(ASValue : string);
begin
  inherited Create;
  SValue := ASValue;
  NValue := StrToFloatDef(ASValue,0);
  ASValue := trim(uppercase(ASValue));
//  if length(ASValue) >= 8 then
//    try DValue := EncodeDate(StrToIntDef(copy(ASValue,1,4),0),StrToIntDef(copy(ASValue,5,2),0),StrToIntDef(copy(ASValue,7,2),0)); except end;
  BValue := (ASValue = 'TRUE') or (ASValue = '-1') or (ASValue = '1') or (ASValue = 'YES') or (ASValue = 'Y');
end;

// ========== TESRIDataSet =====================================================

constructor TESRIDataSet.Create(ASHPFileName : string);
var
  ff : TextFile;
  ss : string;
begin
  inherited Create;
  FSHPFileName    := ASHPFileName;
  FDBFFileName    := ReplaceExtension(ASHPFileName,'DBF');
  FCPGFileName    := ReplaceExtension(ASHPFileName,'CPG');
  FESRIShapes     := nil;
  FESRILines      := nil;
  FESRIPoints     := nil;
  FESRIAttributes := nil;
  FCodePage       := ANSI_CHARSET;
  if not LoadShpFile then
    raise Exception.Create(FSHPFileName+' has an incorrect structure');
  FDBF := LoadDbfFile;
  if FileExists(FCPGFileName) then
  begin
    AssignFile(ff,FCPGFileName);
    reset(ff);
    readln(ff,ss);
    CloseFile(ff);
    FCodePage := GetCPGCodePage(ss);
  end;
end;

// -----------------------------------------------------------------------------

destructor TESRIDataSet.Destroy;
begin
  if assigned(FESRIShapes)     then FreeAndNil(FESRIShapes);
  if assigned(FESRILines)      then FreeAndNil(FESRILines);
  if assigned(FESRIPoints)     then FreeAndNil(FESRIPoints);
  if assigned(FESRIAttributes) then FreeAndNil(FESRIAttributes);
  inherited Destroy;
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.GetVersion : integer;
begin
  result := FHDR.FileVersion;
end;

// -----------------------------------------------------------------------------

function b2l(i : integer) : integer;
begin
  result := (integer(lo(loword(i))) shl 24)+
            (integer(hi(loword(i))) shl 16)+
            (integer(lo(hiword(i))) shl  8)+
            (integer(hi(hiword(i))));
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.GetLength : integer;
begin
  result := b2l(FHDR.FileLength)*2;
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.GetShapeType : integer;
begin
  result := FHDR.ShapeType;
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.GetRect : TRect;
begin
  result.Left   := round(FHDR.Rect.Left * 1000000);
  result.Top    := round(FHDR.Rect.Top * 1000000);
  result.Right  := round(FHDR.Rect.Right * 1000000);
  result.Bottom := round(FHDR.Rect.Bottom * 1000000);
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.LoadShpFile : boolean;
var
  fs         : TFileStream;
  rh         : TESRIRecordHeader;
  ShapeType  : integer;
  PartCount  : integer;
  PointCount : integer;
  Point      : TDoublePoint;
  Parts      : PIntegerArray;
  Points     : PDoublePointArray;
  Rect       : TDoubleRect;
begin
  result := false;
  fs := TFileStream.Create(FSHPFileName,fmOpenRead+fmShareDenyNone);
  try
    if fs.Read(FHDR,sizeof(TESRIFileHeader)) <> sizeof(TESRIFileHeader) then exit;
    repeat
      if fs.Read(rh,sizeof(TESRIRecordHeader)) <> sizeof(TESRIRecordHeader) then exit;
      rh.RecordNumber := b2l(rh.RecordNumber);
      rh.ContentLength := b2l(rh.ContentLength);
      if rh.ContentLength > 0 then
      begin
        if fs.Read(ShapeType,sizeof(integer)) <> sizeof(integer) then exit;
        case ShapeType of
          ESRIPoint:
          begin
            if fs.Read(Point,sizeof(TDoublePoint)) <> sizeof(TDoublePoint) then exit;
            if not assigned(FESRIPoints) then FESRIPoints := TESRIPointList.Create;
            FESRIPoints.AddPoint(rh.RecordNumber,Point);
          end;
          ESRIPolyLine:
          begin
            if fs.Read(Rect,sizeof(TDoubleRect)) <> sizeof(TDoubleRect) then exit;
            if fs.Read(PartCount,sizeof(integer)) <> sizeof(integer) then exit;
            if fs.Read(PointCount,sizeof(integer)) <> sizeof(integer) then exit;
            GetMem(Parts,PartCount*sizeof(integer));
            GetMem(Points,PointCount*sizeof(TDoublePoint));
            try
              if fs.Read(Parts^,PartCount*sizeof(integer)) <> (PartCount*sizeof(integer)) then exit;
              if fs.Read(Points^,PointCount*sizeof(TDoublePoint)) <> (PointCount*sizeof(TDoublePoint)) then exit;
              if not assigned(FESRILines) then FESRILines := TESRILineList.Create;
              FESRILines.AddLine(rh.RecordNumber,PartCount,PointCount,Rect,Parts,Points);
            finally
              FreeMem(Parts);
              FreeMem(Points);
            end;
          end;
          ESRIPolygon:
          begin
            if fs.Read(Rect,sizeof(TDoubleRect)) <> sizeof(TDoubleRect) then exit;
            if fs.Read(PartCount,sizeof(integer)) <> sizeof(integer) then exit;
            if fs.Read(PointCount,sizeof(integer)) <> sizeof(integer) then exit;
            GetMem(Parts,PartCount*sizeof(integer));
            GetMem(Points,PointCount*sizeof(TDoublePoint));
            try
              if fs.Read(Parts^,PartCount*sizeof(integer)) <> (PartCount*sizeof(integer)) then exit;
              if fs.Read(Points^,PointCount*sizeof(TDoublePoint)) <> (PointCount*sizeof(TDoublePoint)) then exit;
              if not assigned(FESRIShapes) then FESRIShapes := TESRIShapeList.Create;
              FESRIShapes.AddShape(rh.RecordNumber,PartCount,PointCount,Rect,Parts,Points);
            finally
              FreeMem(Parts);
              FreeMem(Points);
            end;
          end;
          ESRIMultiPoint:
          begin
            if fs.Read(Rect,sizeof(TDoubleRect)) <> sizeof(TDoubleRect) then exit;
            if fs.Read(PointCount,sizeof(integer)) <> sizeof(integer) then exit;
            GetMem(Points,PointCount*sizeof(TDoublePoint));
            try
              if fs.Read(Points^,PointCount*sizeof(TDoublePoint)) <> (PointCount*sizeof(TDoublePoint)) then exit;
              if not assigned(FESRIPoints) then FESRIPoints := TESRIPointList.Create;
              FESRIPoints.AddMultiplePoints(rh.RecordNumber,PointCount,Points);
            finally
              FreeMem(Points);
            end;
          end;
        end;
      end;
    until fs.Position >= fs.Size;
    result := true;
  finally
    fs.Free;
  end;
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.LoadDbfFile : boolean;
var
  DBF  : TDBF;
  Attr : TESRIAttribute;
  i    : integer;
  v    : TESRIAttributeValue;
begin
  result := false;
  if not FileExists(FDbfFileName) then exit;
  FESRIAttributes := TESRIAttributes.Create;
  try
    DBF := TDBF.Create(FDbfFileName);
  except
    exit;
  end;
  FCodePage := GetDBFCodePage(DBF.Language);
  for i:=0 to DBF.Count-1 do
  begin
    Attr := TESRIAttribute.Create(DBF[i].FieldName,DBF[i].FieldType);
    Attributes.Add(Attr);
  end;
  while not DBF.EOF do
  begin
    for i:=0 to DBF.Count-1 do
      Attributes[i].Add(TESRIAttributeValue.Create(DBF.Fields[i]));
    DBF.Next;
  end;
  result := true;
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.GetDBFCodePage(LanguageCode : byte) : byte;
var
  ii : integer;
begin
  for ii:=1 to MAX_CP do
    if CharsetTable[ii].DbfLanguageCode = LanguageCode then
    begin
      result := CharsetTable[ii].CharsetCode;
      exit;
    end;
  result := ANSI_CHARSET;
end;

// -----------------------------------------------------------------------------

function TESRIDataSet.GetCPGCodePage(CPG : string) : byte;
var
  ii : integer;
begin
  for ii:=1 to MAX_CP do
    if CharsetTable[ii].CharsetString = CPG then
    begin
      result := CharsetTable[ii].CharsetCode;
      exit;
    end;
  result := ANSI_CHARSET;
end;

// ========== TESRIPointList ================================================

procedure TESRIPointList.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    TPointRec(items[i]).Free;
    items[i] := nil;
  end;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

procedure TESRIPointList.AddPoint(AttrIdx : integer; Point : TDoublePoint);
var
  p : TPointRec;
begin
  p := TPointRec.Create;
  p.Attr := AttrIdx;
  p.X := round(Point.X * 1000000);
  p.Y := round(Point.Y * 1000000);
  Add(p);
end;

// -----------------------------------------------------------------------------

procedure TESRIPointList.AddMultiplePoints(Attr,PointCount : integer; Points : PDoublePointArray);
var
  p : TPointRec;
  i : integer;
begin
  for i:=0 to PointCount-1 do
  begin
    p := TPointRec.Create;
    p.Attr := Attr;
    p.X := round(Points[i].X * 1000000);
    p.Y := round(Points[i].Y * 1000000);
    Add(p);
  end;
end;

// -----------------------------------------------------------------------------

function TESRIPointList.GetPointRec(PointIdx : integer) : TPointRec;
begin
  if (PointIdx < 0) or (PointIdx >= Count) then
    result := nil
  else
    result := items[PointIdx];
end;

// ========== TESRILineList ====================================================

procedure TESRILineList.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    TPolyRec(items[i]).Free;
    items[i] := nil;
  end;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

procedure TESRILineList.AddLine(Attr,PartCount,PointCount : integer; Rect : TDoubleRect; Parts : PIntegerArray; Points : PDoublePointArray);
var
  rec : TPolyRec;
  i   : integer;
begin
  rec := TPolyRec.Create;
  rec.Rect.Left   := round(Rect.Left * 1000000);
  rec.Rect.Top    := round(Rect.Top * 1000000);
  rec.Rect.Right  := round(Rect.Right * 1000000);
  rec.Rect.Bottom := round(Rect.Bottom * 1000000);
  rec.Attr := Attr;
  setlength(rec.Parts,PartCount);
  for i:=0 to PartCount-1 do
    rec.Parts[i] := Parts[i];
  setlength(rec.Points,PointCount);
  for i:=0 to PointCount-1 do
  begin
    rec.Points[i].X := round(Points[i].X * 1000000);
    rec.Points[i].Y := round(Points[i].Y * 1000000);
  end;
  Add(rec);
end;

// -----------------------------------------------------------------------------

function TESRILineList.GetPolyCount(LineIdx : integer) : integer;
begin
  if (LineIdx < 0) or (LineIdx >= Count) then
    result := 0
  else
    result := length(TPolyRec(items[LineIdx]).Parts);
end;

// -----------------------------------------------------------------------------

function TESRILineList.GetLine(LineIdx : integer) : TPolyRec;
begin
  if (LineIdx < 0) or (LineIdx >= Count) then
    result := nil
  else
    result := items[LineIdx];
end;

// -----------------------------------------------------------------------------

function TESRILineList.GetPoly(LineIdx,PolyIdx : integer) : TPolygon;
var
  Idx,i,j : integer;
  PolyCount : integer;
  p : TPoint;
begin
  if (LineIdx < 0) or (LineIdx >= Count) then
    setlength(result,0)
  else
  begin
    PolyCount := length(TPolyRec(items[LineIdx]).Parts);
    if (PolyIdx < 0) or (PolyIdx >= PolyCount) then
    begin
      setlength(result,0);
      exit;
    end;
    Idx := TPolyRec(items[LineIdx]).Parts[PolyIdx];
    p := TPolyRec(items[LineIdx]).Points[Idx];
    j := 1;
    for i:=Idx+1 to length(TPolyRec(items[LineIdx]).Points)-1 do
    begin
      inc(j);
      if (p.X = TPolyRec(items[LineIdx]).Points[i].X) and (p.Y = TPolyRec(items[LineIdx]).Points[i].Y) then break;
    end;
    setlength(result,j);
    for i:=Idx to (Idx+j-1) do
    begin
      result[i-Idx] := TPolyRec(items[LineIdx]).Points[i];
    end;
  end;
end;

// -----------------------------------------------------------------------------

function LineByAttributeID(Item1,Item2 : pointer) : integer;
begin
  if TLine(Item1).AttrIdx > TLine(Item2).AttrIdx then
    result := 1
  else
  if TLine(Item1).AttrIdx < TLine(Item2).AttrIdx then
    result := -1
  else
    result := 0;
end;

// -----------------------------------------------------------------------------

procedure TESRILineList.SortByAttribute;
begin
  Sort(LineByAttributeID);
end;

// ========== TESRIShapeList ===================================================

procedure TESRIShapeList.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    TPolyRec(items[i]).Free;
    items[i] := nil;
  end;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

procedure TESRIShapeList.AddShape(Attr,PartCount,PointCount : integer; Rect : TDoubleRect; Parts : PIntegerArray; Points : PDoublePointArray);
var
  rec : TPolyRec;
  i   : integer;
begin
  rec := TPolyRec.Create;
  rec.Rect.Left   := round(Rect.Left * 1000000);
  rec.Rect.Top    := round(Rect.Top * 1000000);
  rec.Rect.Right  := round(Rect.Right * 1000000);
  rec.Rect.Bottom := round(Rect.Bottom * 1000000);
  rec.Attr := Attr;
  setlength(rec.Parts,PartCount);
  for i:=0 to PartCount-1 do
    rec.Parts[i] := Parts[i];
  setlength(rec.Points,PointCount);
  for i:=0 to PointCount-1 do
  begin
    rec.Points[i].X := round(Points[i].X * 1000000);
    rec.Points[i].Y := round(Points[i].Y * 1000000);
  end;
  Add(rec);
end;

// -----------------------------------------------------------------------------

function TESRIShapeList.GetShape(ShapeIdx : integer) : TPolyRec;
begin
  if (ShapeIdx < 0) or (ShapeIdx >= Count) then
    result := nil
  else
    result := items[ShapeIdx];
end;

// -----------------------------------------------------------------------------

function TESRIShapeList.GetPolyCount(ShapeIdx : integer) : integer;
begin
  if (ShapeIdx < 0) or (ShapeIdx >= Count) then
    result := 0
  else
    result := length(TPolyRec(items[ShapeIdx]).Parts);
end;

// -----------------------------------------------------------------------------

function TESRIShapeList.GetPoly(ShapeIdx : integer; PolyIdx : integer) : TPolygon;
var
  Idx,i,j : integer;
  PolyCount : integer;
  p : TPoint;
begin
  if (ShapeIdx < 0) or (ShapeIdx >= Count) then
    setlength(result,0)
  else
  begin
    PolyCount := length(TPolyRec(items[ShapeIdx]).Parts);
    if (PolyIdx < 0) or (PolyIdx >= PolyCount) then
    begin
      setlength(result,0);
      exit;
    end;
    Idx := TPolyRec(items[ShapeIdx]).Parts[PolyIdx];
    p := TPolyRec(items[ShapeIdx]).Points[Idx];
    j := 1;
    for i:=Idx+1 to length(TPolyRec(items[ShapeIdx]).Points)-1 do
    begin
      inc(j);
      if (p.X = TPolyRec(items[ShapeIdx]).Points[i].X) and (p.Y = TPolyRec(items[ShapeIdx]).Points[i].Y) then break;
    end;
    setlength(result,j);
    for i:=Idx to (Idx+j-1) do
    begin
      result[i-Idx] := TPolyRec(items[ShapeIdx]).Points[i];
    end;
  end;
end;

// -----------------------------------------------------------------------------

function TESRIShapeList.GetAttr(ShapeIdx : integer) : integer;
begin
  if (ShapeIdx < 0) or (ShapeIdx >= Count) then
    result := 0
  else
    result := TPolyRec(items[ShapeIdx]).Attr;
end;

// -----------------------------------------------------------------------------

procedure TESRIShapeList.SetAttr(ShapeIdx : integer; Attr : integer);
begin
  if (ShapeIdx >= 0) and (ShapeIdx < Count) then
    TPolyRec(items[ShapeIdx]).Attr := Attr;
end;

// -----------------------------------------------------------------------------

function TESRIShapeList.GetShapeByAttr(Attr : integer) : integer;
var
  ii : integer;
begin
  result := -1;
  if Attr < 0 then exit;
  for ii:=0 to Count-1 do
  begin
    if TPolyRec(items[ii]).Attr = Attr then
    begin
      result := ii;
      exit;
    end;
  end;
end;

// ========== TESRIAttribute ===================================================

constructor TESRIAttribute.Create(AName : string; ADataType : char);
begin
  inherited Create;
  Name     := AName;
  DataType := ADataType;
end;

// -----------------------------------------------------------------------------

procedure TESRIAttribute.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    Values[i].Free;
    items[i] := nil;
  end;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

function TESRIAttribute.GetAttributeValue(Idx : integer) : TESRIAttributeValue;
begin
  if (Idx < 0) or (Idx >= Count) then
    result := nil
  else
    result := items[Idx];
end;

// -----------------------------------------------------------------------------

function TESRIAttribute.GetAttrIdxByValue(Value : string) : integer;
var
  ii : integer;
begin
  result := -1;
  for ii:=0 to Count-1 do
    if Value = TESRIAttributeValue(items[ii]).SValue then
    begin
      result := ii;
      exit;
    end;
end;

// -----------------------------------------------------------------------------

function TESRIAttribute.GetNextAttrIdxByValue(LastIdx : integer; Value : string) : integer;
var
  ii : integer;
begin
  result := -1;
  for ii:=(LastIdx+1) to Count-1 do
    if Value = TESRIAttributeValue(items[ii]).SValue then
    begin
      result := ii;
      exit;
    end;
end;

// ========== TESRIAttributes ==================================================

procedure TESRIAttributes.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    Fields[i].Free;
    items[i] := nil;
  end;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

function TESRIAttributes.GetAttribute(Idx : integer) : TESRIAttribute;
begin
  if (Idx < 0) or (Idx >= Count) then
    result := nil
  else
    result := items[Idx];
end;

// -----------------------------------------------------------------------------

function TESRIAttributes.GetRowCount : integer;
begin
  if Count = 0 then
    result := 0
  else
    result := Fields[0].Count;
end;

// ========== TPolyRec =========================================================

destructor TPolyRec.Destroy;
begin
  setlength(Parts,0);
  setlength(Points,0);
  inherited Destroy;
end;

// =============================================================================

end.
