unit ShapeFileObjects;

interface

uses
  Types, Classes;

type
  TFileFormat = (ffUnknown,ffESRI,ffMapInfo);

type
  TPolygon = array of TPoint;

type
  TShortPoint = packed record
    X,Y : SmallInt;
  end;

type
  TMemPolygon = array [0..$ffff] of TPoint;
  PMemPolygon = ^TMemPolygon;
  TMemShortPoly = array [0..$ffff] of TShortPoint;
  PMemShortPoly = ^TMemShortPoly;

type
  TBoundaryIDArray = array [0..7] of integer;

type
  TShape = class (TList)
    constructor Create(AAttrIdx: integer; ABorderWidth : integer; ABorderColor,AShapeColor : dword);
    procedure Clear; override;
  private
    function GetPolygon(Idx : integer) : TPolygon;
  public
    Attributes  : TBoundaryIDArray;
    BorderWidth : integer;
    BorderColor : dword;
    ShapeColor  : dword;
    R           : TRect;
    property Polygons[Idx : integer] : TPolygon read GetPolygon; default;
    procedure Mirror;
    function GetRect : TRect;
  end;

type
  TShapeSet = class (TList)
    function GetShape(Idx : integer) : TShape;
  public
    property Shapes[Idx : integer] : TShape read GetShape; default;
  end;

type
  TLine = class (TList)
    constructor Create(AAttrIdx : integer; AWidth : integer; AColor : dword);
    procedure Clear; override;
  private
    function GetPoint(Idx : integer) : TPoint;
  public
    AttrIdx : integer;
    Width   : integer;
    Color   : dword;
    property Points[Idx: integer] : TPoint read GetPoint; default;
    procedure AddPoint(X,Y : integer);
    procedure Mirror;
  end;

implementation

uses
  Windows;

// ========== TShape ===========================================================

constructor TShape.Create(AAttrIdx : integer; ABorderWidth : integer; ABorderColor,AShapeColor : dword);
begin
  inherited Create;
  Attributes[0] := AAttrIdx;
  BorderWidth   := ABorderWidth;
  BorderColor   := ABorderColor;
  ShapeColor    := AShapeColor;
  R.Left        := 0;
  R.Top         := 0;
  R.Right       := 0;
  R.Bottom      := 0;
end;

// -----------------------------------------------------------------------------

procedure TShape.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    FreeMem(PMemPolygon(items[i]));
    items[i] := nil;
  end;
  R.Left        := 0;
  R.Top         := 0;
  R.Right       := 0;
  R.Bottom      := 0;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

function TShape.GetPolygon(Idx : integer) : TPolygon;
var
  i,j : integer;
begin
  if (Idx < 0) or (Idx >= Count) then
  begin
    setlength(result,0);
    exit;
  end;
  j := PMemPolygon(items[Idx])[0].X;
  setlength(result,j);
  for i:=0 to j-1 do
  begin
    result[i].X := PMemPolygon(items[Idx])[i+1].X;
    result[i].Y := PMemPolygon(items[Idx])[i+1].Y;
  end;
end;

// -----------------------------------------------------------------------------

procedure TShape.Mirror;
var
  i,j,k : integer;
begin
  for i:=0 to Count-1 do
  begin
    j := PMemPolygon(items[i])[0].X;
    for k:=1 to j do PMemPolygon(items[i])[k].X := -PMemPolygon(items[i])[k].X;
  end;
end;

// -----------------------------------------------------------------------------

function TShape.GetRect : TRect;
var
  i,j : integer;
begin
  if R.Left = 0 then
  begin
    if (Count = 0) or not assigned(items[0]) or (PMemPolygon(items[0])[0].X = 0) then exit;
    R.Left   := PMemPolygon(items[0])[1].X;
    R.Top    := PMemPolygon(items[0])[1].Y;
    R.Right  := PMemPolygon(items[0])[1].X;
    R.Bottom := PMemPolygon(items[0])[1].Y;
    for i:=0 to Count-1 do
    begin
      for j:=1 to PMemPolygon(items[i])[0].X do
      begin
        if R.Left   > PMemPolygon(items[i])[j].X then R.Left   := PMemPolygon(items[i])[j].X;
        if R.Top    < PMemPolygon(items[i])[j].Y then R.Top    := PMemPolygon(items[i])[j].Y;
        if R.Right  < PMemPolygon(items[i])[j].X then R.Right  := PMemPolygon(items[i])[j].X;
        if R.Bottom > PMemPolygon(items[i])[j].Y then R.Bottom := PMemPolygon(items[i])[j].Y;
      end;
    end;
  end;
  result := R;
end;

// ========== TShapeSet ========================================================

function TShapeSet.GetShape(Idx : integer) : TShape;
begin
  if (Idx < 0) or (Idx >= Count) then
    result := nil
  else
    result := items[Idx];
end;

// ========== TLine ============================================================

constructor TLine.Create(AAttrIdx : integer; AWidth : integer; AColor : dword);
begin
  inherited Create;
  AttrIdx := AAttrIdx;
  Width := AWidth;
  Color := AColor
end;

// -----------------------------------------------------------------------------

procedure TLine.Clear;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    FreeMem(PPoint(items[i]));
    items[i] := nil;
  end;
  inherited Clear;
end;

// -----------------------------------------------------------------------------

function TLine.GetPoint(Idx : integer) : TPoint;
begin
  if (Idx < 0) or (Idx >= Count) then
  begin
    result.X := 0;
    result.Y := 0;
  end
  else
    result := PPoint(items[Idx])^;
end;

// -----------------------------------------------------------------------------

procedure TLine.AddPoint(X,Y : integer);
var
  P : PPoint;
begin
  GetMem(P,sizeof(TPoint));
  P.X := X;
  P.Y := Y;
  Add(P);
end;

// -----------------------------------------------------------------------------

procedure TLine.Mirror;
var
  i : integer;
begin
  for i:=0 to Count-1 do
  begin
    PPoint(items[i]).X := -PPoint(items[i]).X;
  end;
end;

// =============================================================================

end.
